import data.*;
import java.io.*;

public class CMain {
  public static void main(String[] args) throws IOException {
    Holder h = Holder.INSTANCE;

    if (args.length < 4) {
      System.out.println("Usage: <c/C> <start size> <end size> <tests per size>");
      return;
    }

    if (args[0].equals("c")) h.dynamicClustering = false;
    else if (args[0].equals("C")) h.dynamicClustering = true;
    else {
      System.out.println("Invalid type argument.");
      return;
    }

    int a = Integer.parseInt(args[1]);
    int b = Integer.parseInt(args[2]);
    int m = Integer.parseInt(args[3]);

    String fname = "plot/"+args[0]+"_"+a+"_"+b+"_"+m;
    String tfname = "plot/t"+args[0]+"_"+a+"_"+b+"_"+m;
    PrintWriter out = new PrintWriter(new File(fname+"-all.dat"));
    PrintWriter mout = new PrintWriter(new File(fname+"-med.dat"));
    PrintWriter tout = new PrintWriter(new File(tfname+"-all.dat"));
    PrintWriter tmout = new PrintWriter(new File(tfname+"-med.dat"));

    //warmup
    for(int i=0; i<200; ++i) {
      double x = 0, y = Math.random();
      h.labels.add(new Label("",x,y));
      h.generate();
    }

    double pta=1e9;
    int skip=0;
    for(int i=a; i<=b; ++i) {
      int s=0;
      long ts=0;

//      System.gc();
      for(int k=0; k<m; ++k) {
        h.reset();
        for(int j=0; j<i; ++j) {
          double x = 0, y = Math.random();
          h.labels.add(new Label("",x,y));
        }

        long t0, tt;
        int ss=0;
        while(true) {
          t0 = System.nanoTime();

//          System.out.println("testcase starting");
          h.generate();

          tt = System.nanoTime() - t0;

          if (i>a && tt > 2.*(pta+5e6) && ss<5) {
            System.out.println("skip "+(++skip));
            System.out.println(tt/1e6+" "+pta/1e6);
            ++ss;
            continue;
          }
          break;
        }
        s += h.positioner.complexity;
        ts += tt;
        tout.printf("%d %f\n", i, tt/1e6);
        out.printf("%d %d\n", i, h.positioner.complexity);
      }
      double avg = (double)s / m;
      double tavg = (double)ts / m;
      System.out.printf("%d %f %f\n", i, avg, tavg/1e6);
      mout.printf("%d %f\n", i, avg);
      tmout.printf("%d %f\n", i, tavg/1e6);
      pta = tavg;
      if (!h.dynamicClustering) {
        if (i%2==0) pta *= .5;
        else pta *= 2;
      }
    }
    out.close();
    mout.close();
    tout.close();
    tmout.close();
    System.out.println("skipped "+skip+" cases");
  }
}
