package data;

import java.util.*;

/**
 * Purely functional balanced binary tree map.
 */
public class AASet<Key extends Comparable<Key>> implements Iterable<Key>, BinaryTree<Key> {
  private AASet<Key> left, right;
  private int level;

  private Key key;

  public AASet() {
    left = this;
    right = this;
  }
  public AASet(Key k) {
    left = new AASet<Key>();
    right = new AASet<Key>();
    key = k;
    level = 1;
  }
  private AASet(Key k, AASet<Key> l, AASet<Key> r, int lev) {
    key = k;
    left = l;
    right = r;
    level = lev;
  }

  public boolean isEmpty() {
    return level==0;
  }
  public AASet<Key> add(Key k) {
    if (level==0) return new AASet<Key>(k);
    AASet<Key> t = new AASet<Key>(key, left, right, level);
    int cmp = k.compareTo(key);
    if (cmp < 0) {
      t.left = t.left.add(k);
    } else if (cmp > 0) {
      t.right = t.right.add(k);
    } else {
      t.key = k;
      return t;
    }
    t = t.skew();
    t = t.split();
    return t;
  }
  /*
  public Value get(Key k) {
    if (level==0) return null;
    int cmp = k.compareTo(key);
    if (cmp < 0) return left.get(k);
    if (cmp > 0) return right.get(k);
    return value;
  }*/
  public AASet<Key> remove(Key k) {
    if (level==0) return this;
    AASet<Key> t = new AASet<Key>(key, left, right, level);
    int cmp = k.compareTo(key);
    if (cmp < 0) {
      t.left = t.left.remove(k);
    } else if (cmp > 0) {
      t.right = t.right.remove(k);
    } else {
      if (right.level==0) return new AASet<Key>();
      if (left.level==0) {
        AASet<Key> l = t.successor();
        t.right = t.right.remove(l.key);
        t.key = l.key;
      } else {
        AASet<Key> l = t.predecessor();
        t.left = t.left.remove(l.key);
        t.key = l.key;
      }
    }

    t.decreaseLevel();
    t = t.skew();
    t.right = t.right.skew();
    t.right.right = t.right.right.skew();
    t = t.split();
    t.right = t.right.split();
    return t;
  }

  private AASet<Key> skew() {
    AASet<Key> l = left;
    if (level!=0 && level == l.level) {
      left = l.right;
      l.right = this;
      return l;
    }
    return this;
  }
  private AASet<Key> split() {
    AASet<Key> r = right;
    if (level!=0 && level==r.right.level) {
      right = r.left;
      r.left = this;
      ++r.level;
      return r;
    }
    return this;
  }
  private AASet<Key> successor() {
    AASet<Key> t = right;
    while(t.left.level > 0) t = t.left;
    return t;
  }
  private AASet<Key> predecessor() {
    AASet<Key> t = left;
    while(t.right.level > 0) t = t.right;
    return t;
  }
  private void decreaseLevel() {
    int s = Math.min(left.level, right.level)+1;
    if (s < level) {
      level = s;
      if (s < right.level)
        right.level = s;
    }
  }

  public Iterator<Key> iterator() {
    if (level==0) return new BinaryTreeIterator<Key>();
    return new BinaryTreeIterator<Key>(this);
  }
  public Key getValue() {
    return key;
  }
  public AASet<Key> getLeft() {
    if (left.level==0) return null;
    return left;
  }
  public AASet<Key> getRight() {
    if (right.level==0) return null;
    return right;
  }

  public String toString() {
    StringBuffer r = new StringBuffer("{");
    for(Key k : this)
      r.append(k.toString()+",");
    r.append("}");
    return r.toString();
  }

  public int getLevel() {
    return level;
  }
}
