package data;

import gui.Listeners;
import gui.WindowInterface;
import gui.ZoomAndPanListener;
import gui.LabelingPanel;

import java.util.Set;
import java.util.TreeSet;
import java.util.TreeMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Collections;

import java.awt.geom.*;
import java.awt.*;

//public class Holder {
//  public static Holder INSTANCE = new Holder();
public enum Holder {
  INSTANCE;

  public Listeners listener = new Listeners();

  // TreeSet takes care of the sorting of the labels for us
  public TreeSet<Label> labels = new TreeSet<Label>();
  public TreeSet<Label> allLabels = new TreeSet<Label>();
  public List<Label> lowPriority = new ArrayList<Label>();

  public LabelingPanel window;
  private double zoomPercentage;

  public boolean dynamicClustering = true;
  public boolean onlyVisible = true;
  public boolean keepOnScreen = true;

  public int maxVisible = 10;

  public LabelPositioner positioner = new LabelPositioner();

  int nextPriority = -1;

  Holder() {
    Label[] lb = new Label[]{
        new Label("1", 231.000000, 29.000000),
        new Label("2", 81.000000, 80.000000),
        new Label("4", 383.000000, 115.000000),
        new Label("5", 135.567047, 143.199478),
        new Label("6", 348.183777, 208.522507),
        new Label("3", 271.000000, 231.000000),
        new Label("11", 467.655121, 279.001251),
        new Label("9", 109.074188, 309.450378),
        new Label("7", 443.053772, 330.004059),
        new Label("10", 240.451614, 404.606964),
        new Label("8", 97.849991, 469.787292),
        new Label("12", 370.657715, 490.864014)
    };
    lb = sampleCase;
    lb = sampleCase2;
    for(int i=0; i<lb.length; ++i) {
//      lb[i].y = 1538 - lb[i].y;
      lb[i].text = "Label "+lb[i].text;
      lb[i].priority = i;
      labels.add(lb[i]);
    }
    allLabels = new TreeSet<Label>(labels);
    generate();
  }
  public Dimension generateRandom(int size) {
    labels.clear();
    double ss = Math.sqrt((double)size);
    double ws = 150*ss;
    double hs = 120*ss;
    for (int i = 0; i < size; i++) {
      labels.add(DataGenerator.getLabel(ws, hs, i));
//      labels.add(DataGenerator.getLabel(360, 240, i));
    }
    allLabels = new TreeSet<Label>(labels);
    generate();
    window.labelsChanged();
    window.repaint();
    return new Dimension((int)ws, (int)hs);
  }

  public void addLabel(Label l) {
    allLabels.add(l);
    generate();
  }
  
  public void reset() {
    labels = new TreeSet<Label>();
    allLabels = new TreeSet<Label>();
    if (window!=null) window.repaint();
    generate();
  }
  public void generate() {
    positioner.generate(labels, dynamicClustering);
  }
  public boolean checkVisible(AffineTransform at, Rectangle bounds, FontMetrics metric, boolean internalLabels) {
    int add = 2;
    bounds.x -= add;
    bounds.y -= add;
    bounds.width += 2*add;
    bounds.height += 2*add;
    int limit = maxVisible<=0 ? allLabels.size() : maxVisible;
    if (onlyVisible) {
      TreeSet<Label> s = new TreeSet<Label>();
      Point2D p = new Point2D.Double();
      for(Label l : allLabels) {
        at.transform(l, p);
//        if (bounds.contains(l))
        if (bounds.contains(p))
          s.add(l);
        else if (internalLabels) {
          double x1 = p.getX(), y1 = p.getY();
          double x2 = x1 + metric.stringWidth(l.text) * at.getScaleX();
          double y2 = y1 + metric.getHeight() * at.getScaleY();
    
          if (x1 <= bounds.x+bounds.width
              && x2 >= bounds.x
              && y1 <= bounds.y+bounds.height
              && y2 >= bounds.y)
          {
            s.add(l);
          }
        }
      }
      if (s.size() > limit) {
        ArrayList<Label> lst = new ArrayList<Label>(s);
        Collections.sort(lst, new Label.PriorityCmp());
        s = new TreeSet<Label>(lst.subList(0, limit));
        lowPriority = lst.subList(limit, lst.size());
      } else lowPriority = new ArrayList<Label>();
      if (!s.equals(labels)) {
        labels = s;
        generate();
        return true;
      }
    } else if (labels!=allLabels) {
      labels = allLabels;
      generate();
      return true;
    }
    return false;
  }
  
  public void setZoomLevel(double zoomLevel) {
    setZoomPercentage(zoomLevel / ZoomAndPanListener.MAX_ZOOM_LEVEL);
  }
  
  private void setZoomPercentage(double zp) {
    if(zp != zoomPercentage) {
      zoomPercentage = zp;
      if (window!=null) window.repaint();
    }
  }
  
  public double getZoomPercentage() {
    return zoomPercentage;
  }
  
  public void printContents() {
    for(Label l: labels) {
      System.out.println("(" + l.x + ", " + l.y + ")");
    }
  }

  private double zoomFactor;
  public void setHeights(int rectSize, double zoom, double ymin, double ymax) {
    zoom *= rectSize;
    positioner.setHeights(zoom, ymin, ymax);
    zoomFactor = zoom;
  }

  public double getZoomFactor() {
    return zoomFactor;
  }
  public TreeMap<Double,Integer> getMedianLine() {
    return positioner.medianStarting;
  }
  public TreeMap<Double,Double> getHeights() {
    return positioner.heights;
  }
  public List<Line2D.Double> getClusterLines() {
    return positioner.clusterLines;
  }

  private Label[] sampleCase = new Label[] {
    new Label("CUBA", 686.1611, 622.584),
    new Label("COLOMBIA", 721.2964, 775.4404),
    new Label("PERU", 712.4409, 889.8428),
    new Label("BOLIVIA", 799.4165, 953.4932),
    new Label("CHILE", 750.5278, 1080.3633),
    new Label("PAPUA", 2397.457, 851.2529),
    new Label("NEW GUINEA", 2383.4863, 862.9941),
    new Label("ZEALAND", 2485.0068, 1151.5742),
    new Label("C A N A D A", 598.4639, 326.7354),
    new Label("MEXICO", 523.2236, 613.728),
    new Label("BELIZE", 635.7617, 653.5439),
    new Label("HONDURAS", 634.5361, 672.9121),
    new Label("NICARAGUA", 647.0649, 695.8804),
    new Label("EL SALVADOR", 586.2256, 698.2559),
    new Label("GUATEMALA", 586.2256, 672.8398),
    new Label("COSTA", 634.3589, 727.7139),
    new Label("RICA", 637.8716, 734.7139),
    new Label("PANAMA", 678.4556, 739.2246),
    new Label("C H I N A", 2002.3164, 510.4805),
    new Label("SURINAME", 848.6675, 770.5439),
    new Label("BAHAMAS", 736.9204, 584.0635),
    new Label("THE", 745.8481, 575.4966),
    new Label("THE GAMBIA", 1126.4077, 697.0928),
    new Label("SWAZILAND", 1531.1484, 1047.8135),
    new Label("ZIMBABWE", 1502.9277, 973.2236),
    new Label("ZAMBIA", 1487.3066, 928.5142),
    new Label("ANGOLA", 1415.7368, 913.6826),
    new Label("COMOROS", 1631.0879, 913.6826),
    new Label("SEYCHELLES", 1674.2871, 877.5332),
    new Label("TANZANIA", 1542.7266, 870.3232),
    new Label("AFRICA", 1459.4375, 1075.8232),
    new Label("SOUTH", 1460.0879, 1061.9219),
    new Label("KENYA", 1577.0176, 802.083),
    new Label("UGANDA", 1541.668, 790.4873),
    new Label("AFRICAN REPUBLIC", 1422.7173, 753.9839),
    new Label("CENTRAL", 1440.5073, 744.6235),
    new Label("YEMEN", 1650.3789, 668.1719),
    new Label("N I G E R", 1347.478, 666.2153),
    new Label("LIBERIA", 1211.9077, 762.1201),
    new Label("CAMEROON", 1369.9077, 759.3564),
    new Label("GABON", 1379.2378, 814.9722),
    new Label("EQUATORIAL GUINEA", 1274.3071, 778.8228),
    new Label("GUINEA", 1203.4077, 713.376),
    new Label("GUINEA", 1133.4976, 714.0244),
    new Label("M A L I", 1274.3979, 650.9517),
    new Label("MAURITANIA", 1188.2183, 641.3032),
    new Label("SENEGAL", 1175.8276, 681.6963),
    new Label("CAPE VERDE", 1089.3579, 669.6733),
    new Label("NORWAY", 1354.2476, 249.8408),
    new Label("SWEDEN", 1403.6372, 261.0723),
    new Label("FINLAND", 1460.2285, 272.8091),
    new Label("FRANCE", 1310.9771, 400.897),
    new Label("SPAIN", 1268.2769, 465.0479),
    new Label("EST", 1469.2285, 305.1372),
    new Label("LAT", 1472.5469, 319.6802),
    new Label("LITH", 1458.3184, 327.4307),
    new Label("ITALY", 1383.4771, 435.313),
    new Label("TUNISIA", 1366.6382, 504.793),
    new Label("LAOS", 2059.9863, 633.168),
    new Label("JAPAN", 2301.627, 481.752),
    new Label("PHILIPPINES", 2246.1875, 693.0723),
    new Label("MARSHALL", 2602.7285, 711.8643),
    new Label("ISLANDS", 2608.5586, 723.96),
    new Label("SOLOMON", 2500.5566, 862.7725),
    new Label("ISLANDS", 2519.7168, 873.1426),
    new Label("FIJI", 2626.0977, 945.9639),
    new Label("TONGA", 2690.8574, 995.9043),
    new Label("THAILAND", 2051.207, 670.4648),
    new Label("BANGLADESH", 1938.957, 594.5039),
    new Label("CAMBODIA", 2082.0977, 695.8076),
    new Label("VIETNAM", 2118.168, 676.8721),
    new Label("MALDIVES", 1812.166, 778.8975),
    new Label("SRI", 1927.1484, 752.5459),
    new Label("LANKA", 1922.1875, 760.1748),
    new Label("SINGAPORE", 2095.5566, 801.5723),
    new Label("BRUNEI", 2147.8359, 765.6494),
    new Label("EAST", 2264.4863, 886.3623),
    new Label("TIMOR", 2262.4766, 893.3633),
    new Label("NEW", 2494.5879, 1139.4033),
    new Label("NORTH KOREA", 2190.8867, 457.7754),
    new Label("MOROCCO", 1213.5571, 541.6563),
    new Label("TRINIDAD AND", 837.937, 716.688),
    new Label("TOBAGO", 844.1274, 722.5918),
    new Label("REPUBLIC", 766.0776, 638.2803),
    new Label("DOMINICAN", 763.3472, 631.4409),
    new Label("JAMAICA", 683.728, 649.1401),
    new Label("PARAGUAY", 841.1782, 1001.8828),
    new Label("ICELAND", 1198.228, 256.4653),
    new Label("UNITED", 1280.8081, 318.9595),
    new Label("KINGDOM", 1277.7183, 357.1201),
    new Label("IRELAND", 1239.9067, 354.0967),
    new Label("LIECH", 1372.3179, 392.4731),
    new Label("SWITZ", 1356.4077, 396.8643),
    new Label("LUX", 1338.7671, 385.2729),
    new Label("CZECH REP", 1400.0376, 380.3047),
    new Label("SLOVAKIA", 1432.438, 384.335),
    new Label("ROMANIA", 1463.6172, 407.375),
    new Label("ALB", 1431.7173, 447.4082),
    new Label("BELARUS", 1474.6367, 350.9277),
    new Label("SYRIA", 1571.7578, 503.7842),
    new Label("CYPRUS", 1521.8672, 509.1133),
    new Label("LEB", 1552.0273, 507.8887),
    new Label("UZBEKISTAN", 1726.418, 439.1279),
    new Label("U K R A I N E", 1488.5273, 379.2241),
    new Label("MOLDOVA", 1494.2871, 391.6074),
    new Label("GEORGIA", 1578.1973, 437.1895),
    new Label("AZERBAIJAN", 1628.998, 453.2393),
    new Label("ARMENIA", 1602.1465, 454.1045),
    new Label("U N I T E D     S T A T E S", 498.7451, 483.3359),
    new Label("I R A N", 1683.0781, 532.2959),
    new Label("IRAQ", 1607.0371, 511.4883),
    new Label("AFGHANISTAN", 1750.4082, 517.0322),
    new Label("PAKISTAN", 1774.0078, 564.6963),
    new Label("ANDORRA", 1293.8374, 441.2158),
    new Label("MARINO", 1373.187, 424.5112),
    new Label("SAN", 1377.147, 419.9043),
    new Label("AUSTRIA", 1398.3169, 397.4102),
    new Label("HUNGARY", 1425.0972, 403.2725),
    new Label("SER", 1444.7769, 425.8232),
    new Label("MONT", 1444.7769, 431.3232),
    new Label("CROATIA", 1420.1978, 413.1367),
    new Label("SLOVENIA", 1379.1577, 409.3213),
    new Label("BURMA", 2007.0762, 622.7295),
    new Label("I N D I A", 1871.3477, 634.9678),
    new Label("NEPAL", 1902.6777, 559.0088),
    new Label("BHUTAN", 1963.2285, 557.7109),
    new Label("T U R K E Y", 1532.5977, 471.0225),
    new Label("NAURU", 2580.207, 814.0332),
    new Label("B R A Z I L", 866.5981, 895.603),
    new Label("A L G E R I A", 1285.7769, 559.728),
    new Label("L I B Y A", 1410.0474, 571.3203),
    new Label("E G Y P T", 1495.5879, 579.8877),
    new Label("BOS", 1414.0073, 422.7119),
    new Label("NIGERIA", 1345.0366, 718.8496),
    new Label("SAMOA", 2706.5781, 914.1338),
    new Label("VATICAN", 1375.4888, 441.5049),
    new Label("MALTA", 1388.978, 492.3364),
    new Label("NETH", 1323.2876, 357.8398),
    new Label("BELGIUM", 1321.9976, 380.1611),
    new Label("DENMARK", 1360.9478, 324.0723),
    new Label("JORDAN", 1563.3379, 542.9521),
    new Label("ISRAEL", 1538.2773, 529.9922),
    new Label("KUWAIT", 1658.9473, 551.2324),
    new Label("BAHRAIN", 1645.4883, 575.2803),
    new Label("QATAR", 1661.8281, 598.5361),
    new Label("UNITED ARAB", 1691.498, 600.0479),
    new Label("EMIRATES", 1697.3281, 606.0225),
    new Label("OMAN", 1719.3574, 624.3115),
    new Label("OMAN", 1713.2383, 578.3037),
    new Label("GERMANY", 1362.1675, 369.1455),
    new Label("POLAND", 1425.2378, 348.7681),
    new Label("ST", 791.2974, 686.5234),
    new Label("THE GRENADINES", 791.228, 691.4297),
    new Label("GRENADA", 833.2583, 704.0879),
    new Label("BARBADOS", 850.6079, 692.9277),
    new Label("ST", 840.6675, 685.0801),
    new Label("DOMINICA", 838.9468, 671.5444),
    new Label("ANTIGUA AND BARBUDA", 836.9272, 655.417),
    new Label("ST", 834.9868, 645.2637),
    new Label("M O N G O L I A", 1989.4277, 408.3838),
    new Label("DJIBOUTI", 1625.5977, 701.4639),
    new Label("BOTSWANA", 1462.3867, 1000.2231),
    new Label("NAMIBIA", 1414.147, 983.0142),
    new Label("TURKMENISTAN", 1690.418, 463.9678),
    new Label("KYRGYZSTAN", 1821.2383, 450.2168),
    new Label("TAJIKISTAN", 1818.8672, 477.1455),
    new Label("BULGARIA", 1472.2578, 439.9209),
    new Label("RUSSIA", 1423.1675, 331.7412),
    new Label("GUYANA", 834.9878, 756.5049),
    new Label("URUGUAY", 883.228, 1093.3936),
    new Label("ECUADOR", 688.2495, 818.3525),
    new Label("BENIN", 1311.7671, 716.6162),
    new Label("TOGO", 1306.2983, 729.8633),
    new Label("D", 1248.8374, 735.1201),
    new Label("CÔTE", 1254.2378, 728.6392),
    new Label("SAUDI", 1598.7578, 576.5029),
    new Label("ARABIA", 1627.418, 612.3604),
    new Label("SIERRA", 1185.8374, 739.4404),
    new Label("LEONE", 1187.6382, 746.7114),
    new Label("ERITREA", 1582.0566, 667.1523),
    new Label("SOUTH", 2214.4375, 487.3828),
    new Label("KOREA", 2214.7969, 494.7246),
    new Label("CITY", 1379.5181, 445.3921),
    new Label("GREECE", 1448.4272, 469.7998),
    new Label("MACEDONIA", 1455.0469, 452.4453),
    new Label("AUSTRALIA", 62.3525, 70.4165),
    new Label("MAURITIUS", 1736.2773, 986.9717),
    new Label("MADAGASCAR", 1632.0273, 989.1318),
    new Label("PALAU", 2326.6074, 757.1514),
    new Label("VANUATU", 2573.8574, 951.2637),
    new Label("TUVALU", 2633.9883, 890.2432),
    new Label("BURKINA", 1279.6567, 688.6807),
    new Label("FASO", 1285.3472, 694.3682),
    new Label("ARGENTINA", 813.8169, 1122.1924),
    new Label("PORTUGAL", 1219.7573, 457.7041),
    new Label("MONACO", 1335.4575, 426.2402),
    new Label("LESOTHO", 1489.5371, 1050.2637),
    new Label("CHAD", 1427.1177, 676.8711),
    new Label("VENEZUELA", 779.8374, 744.7676),
    new Label("GHANA", 1283.3276, 736.3447),
    new Label("S U D A N", 1495.5879, 700.7764),
    new Label("RWANDA", 1502.0664, 821.2334),
    new Label("BURUNDI", 1529.9258, 847.4424),
    new Label("MALAWI", 1566.2871, 934.2744),
    new Label("MOZAMBIQUE", 1553.6875, 962.3525),
    new Label("ETHIOPIA", 1585.0781, 738.792),
    new Label("SOMALIA", 1653.1875, 759.96),
    new Label("FEDERATED STATES OF MICRONESIA", 2377.0781, 717.0479),
    new Label("HAITI", 744.3364, 654.1201),
    new Label("CONGO", 1412.2778, 819.8628),
    new Label("REP", 1411.4175, 803.0132),
    new Label("THE", 1416.8188, 811.7314),
    new Label("OF THE CONGO", 1442.6675, 847.8037),
    new Label("DEMOCRATIC", 1446.6978, 813.0229),
    new Label("REPUBLIC", 1454.8262, 829.5142),
    new Label("ANGOLA", 1370.4478, 853.0039),
    new Label("SAO TOME AND PRINCIPE", 1267.1675, 796.5342),
    new Label("K A Z A K H S T A N", 1716.0586, 390.541),
    new Label("TRISTAN DA CUNHA", 1200.7373, 1142.2637),
    new Label("QUEEN ELIZABETH", 796.5371, 164.0156),
    new Label("ISLANDS", 815.2588, 178.0566),
    new Label("KURIL", 2348.4961, 390.3135),
    new Label("ISLANDS", 2344.3184, 397.8018),
    new Label("NEW SIBERIAN ISLANDS", 2017.4375, 169.1289),
    new Label("SEVERNAYA", 1804.9668, 141.4082),
    new Label("ZEMLYA", 1810.4375, 149.2568),
    new Label("FRANZ JOSEF", 1561.6074, 131.5439),
    new Label("LAND", 1572.2656, 139.1035),
    new Label("NOVAYA", 1587.0273, 172.7998),
    new Label("ZEMLYA", 1586.5977, 179.4248),
    new Label("ALEUTIAN ISLANDS", 2472.8379, 346.3213),
    new Label("PENEDOS DE", 1088.3579, 786.3125),
    new Label("SÃO PEDRO E SÃO PAULO", 1073.5981, 793.0796),
    new Label("FERNANDO DE NORONHA", 1008.9878, 829.083),
    new Label("GALAPAGOS", 562.0513, 815.832),
    new Label("ISLANDS", 568.1704, 823.1025),
    new Label("HAWAIIAN", 79.9209, 628.9922),
    new Label("ISLANDS", 82.0801, 637.7764),
    new Label("ÎLES MARQUISES", 243.5654, 878.5435),
    new Label("SOCIETY", 132.6953, 959.4727),
    new Label("ISLANDS", 133.6934, 966.2534),
    new Label("ÎLES TUBUAI", 189.5039, 1016.7832),
    new Label("ARCHIPEL DES TUAMOTU", 250.0576, 942.334),
    new Label("ARCHIPIELAGO", 703.2974, 1105.3438),
    new Label("JUAN FERNANDEZ", 698.7427, 1112.4443),
    new Label("SOUTH ORKNEY", 1046.959, 1329.1943),
    new Label("ISLANDS", 1055.5991, 1336.7529),
    new Label("PRINCE EDWARD", 1535.2305, 1218.0938),
    new Label("ISLANDS", 1546.3213, 1225.4336),
    new Label("ILES CROZET", 1634.2598, 1215.3633),
    new Label("ILES KERGUELEN", 1718.2109, 1239.4834),
    new Label("CHATHAM ISLANDS", 2543.7617, 1195.2031),
    new Label("BOUNTY ISLANDS", 2517.9922, 1221.043),
    new Label("SNARES ISLANDS", 2428.4316, 1222.5918),
    new Label("ANTIPODES ISLANDS", 2499.7012, 1237.1729),
    new Label("AUCKLAND", 2375.0723, 1237.0322),
    new Label("ISLANDS", 2375.9316, 1243.873),
    new Label("KERMADEC", 2637.6523, 1072.4414),
    new Label("ISLANDS", 2640.1035, 1079.7822),
    new Label("ISLANDS", 2011.6777, 695.8799),
    new Label("SPRATLY", 2138.5078, 725.54),
    new Label("NICOBAR", 1988.9277, 743.5439),
    new Label("PARACEL", 2155.6074, 660.3115),
    new Label("ISLANDS", 1989.7168, 748.8066),
    new Label("ISLANDS", 2137.9277, 731.876),
    new Label("ISLANDS", 2154.8164, 666.8643),
    new Label("AZORES", 1095.187, 478.8003),
    new Label("CANARY ISLANDS", 1166.4771, 549.8633),
    new Label("MADEIRA", 1156.8979, 509.9751),
    new Label("AZORES", 91.2969, 110.3037),
    new Label("ISLAS", 446.8828, 649.7959),
    new Label("REVILLAGIGEDO", 435.1699, 655.6323),
    new Label("ANDAMAN", 2008.8789, 690.0479),
    new Label("LAKSHADWEEP", 1804.3477, 715.6748),
    new Label("ARQUIPELAGO DE", 1019.3491, 822.603),
    new Label("BALEARIC", 1331.3481, 464.4722),
    new Label("ISLANDS", 1332.3579, 469.0796)
  };
  Label[] sampleCase2 = new Label[] {
    new Label("15", 285.202820, 478.420441),
    new Label("14", 315.113037, 599.408020),
    new Label("8", 743.064758, 169.686661),
    new Label("22", 743.286377, 590.506897),
    new Label("21", 709.752502, 598.320557),
    new Label("12", 714.891052, 417.129211),
    new Label("1", 149.753082, 181.577164),
    new Label("11", 773.691101, 405.579224),
    new Label("5", 470.516418, 98.346809),
    new Label("4", 203.223389, 154.503586),
    new Label("2", 249.248474, 117.954254),
    new Label("20", 664.155884, 606.704102),
    new Label("10", 516.441040, 361.479187),
    new Label("3", 90.868065, 91.557518),
    new Label("19", 209.442627, 445.591003),
    new Label("18", 259.949402, 373.366302),
    new Label("7", 396.605560, 47.177750),
    new Label("6", 558.640930, 84.843864),
    new Label("17", 350.356567, 404.175446),
    new Label("23", 727.984497, 572.274902),
    new Label("9", 695.743774, 456.645569),
    new Label("25", 797.637939, 578.481995),
    new Label("24", 700.962097, 554.694031),
    new Label("13", 141.626648, 647.462219),
    new Label("16", 255.403809, 468.319061),
  };
}
