package data;

import java.util.*;

/**
 * Robust kinetic priority queue.
 */
public class KPQ {

  int minSlope;
  double minY;

  double next;

  KPQ c1, c2;
  int size;

  private HashSet<Integer> members = new HashSet<Integer>();

  KPQ() {
  }
  KPQ(int slope, double y0) {
    add(slope, y0, 0);
  }

  void addAll(KPQ q, double t) {
//    System.out.println("lol: "+this);
    if (q.size==0) return;
    if (q.size==1) {
      add(q.minSlope, q.minY, t);
    } else {
      addAll(q.c1, t);
      addAll(q.c2, t);
    }
  }

  void addTo(Collection<Integer> c, int f) {
//    System.out.println("addto "+this);
    if (size==0) return;
    if (size==1) {
      c.add(minSlope*f);
    } else {
      c1.addTo(c, f);
      c2.addTo(c, f);
    }
  }

  void add(int s, double y, double t) {
    ++size;
    members.add(s);
    if (size==1) {
      minSlope = s;
      minY = y;
      next = 1e100;
      return;
    }
    if (c1==null) {
      c1 = new KPQ(minSlope, minY);
      c2 = new KPQ(s, y);
    } else {
      KPQ c = c1.size <= c2.size ? c1 : c2;
      c.add(s, y, t);
    }
    KPQ c = c1.height(t) <= c2.height(t) ? c1 : c2;
//    System.out.println("Comparing @ "+t+" : "+c1.minSlope+","+c1.minY+" "+c2.minSlope+","+c2.minY+" : "+c1.height(t)+" "+c2.height(t));
    minSlope = c.minSlope;
    minY = c.minY;
    setNext();
  }
  void remove(int s, double t) {
//    System.out.println("removing "+s+" "+t+"  from "+this+" ; "+size);
    --size;
    members.remove(s);
    if (size==0) return;

    KPQ c = c1.members.contains(s) ? c1 : c2;
    boolean lower = minSlope == c.minSlope;
    c.remove(s, t);
    if (c.size==0) {
      KPQ d = c==c1 ? c2 : c1;
      members = d.members;
      minSlope = d.minSlope;
      minY = d.minY;
      c1 = d.c1;
      c2 = d.c2;
      size = d.size;
      next = d.next;
      return;
    }
    if (lower) {
      KPQ d = c1.height(t) < c2.height(t) ? c1 : c2;
      minSlope = d.minSlope;
      minY = d.minY;
//      System.out.println("chose new minslope from "+c1.minSlope+" and "+c2.minSlope+" : "+minSlope);
    }
    if (size==1) {
      KPQ d = c1.size>0 ? c1 : c2;
      minSlope = d.minSlope;
      minY = d.minY;
      c1 = null;
      c2 = null;
    }
    setNext();
  }
  boolean advance() {
    assert size > 1;
    double t = intersectionT();
    if (c1.next < t || c2.next < t || minSlope==Math.min(c1.minSlope,c2.minSlope)) {
      KPQ c = c1.next < c2.next ? c1 : c2;
      KPQ d = c==c1 ? c2 : c1;
      t = c.next;
      boolean lower = minSlope == c.minSlope;
      c.advance();
      if (lower) {
        minSlope = c.minSlope;
        minY = c.minY;
        setNext();
        return true;
      }
      setNext();
      return false;
    } else {
      KPQ c = c1.minSlope < c2.minSlope ? c1 : c2;
      minSlope = c.minSlope;
      minY = c.minY;
      next = Math.min(c1.next, c2.next);
      return true;
    }
  }

  private double intersectionT() {
    return (c2.minY - c1.minY) / (c1.minSlope - c2.minSlope);
  }
  private void setNext() {
    if (size==1) {
      next=1e100;
    } else {
      next = Math.min(c1.next, c2.next);
      if (minSlope != Math.min(c1.minSlope, c2.minSlope))
        next = Math.min(next, intersectionT());
    }
  }
  private double height(double t) {
    return minY + t*minSlope;
  }

  public String toString() {
    if (size==0) return "[]";
    if (size==1) return "("+minSlope+","+minY+")";
    return "["+c1+" ; "+c2+"]";
  }
}
