package data;

import java.util.*;
import java.awt.geom.*;

public class LabelPositioner {

  public List<Label> points;
  TreeMap<Double,Integer> medianStarting = new TreeMap<Double,Integer>();
  TreeMap<Double,Integer> lowerStarting = new TreeMap<Double,Integer>();
  TreeMap<Double,Double> heights = new TreeMap<Double,Double>();

  public class ClusterInfo implements Comparable<ClusterInfo> {
//    public List<Integer> points;
    public AASet<Integer> points;
    public int med;
    public int med2;

    ClusterInfo(AASet<Integer> p, int m, int m2) {
      points = p;
      med = m;
      med2 = m2;
    }

    public String toString() {
      return "[Cluster "+med+","+med2+" : "+points+"]";
    }
    public int compareTo(ClusterInfo c) {
      return med - c.med;
    }
  }
  public TreeMap<Double,AAMap<Integer,ClusterInfo>> clustering = new TreeMap<Double,AAMap<Integer,ClusterInfo>>();
  public ArrayList<Line2D.Double> clusterLines;

  public int complexity;

  int getMedianElement(double t) {
    return medianStarting.floorEntry(t).getValue();
  }
  double getHeight(int n, double t) {
    return points.get(n).y - t*(n+1);
  }

  private double limitHeight(int idx, double y, double ymin, double ymax, double t) {
    if (!Holder.INSTANCE.keepOnScreen) return y;
    if (points.size()*t > ymax-ymin) return y;
    y = Math.max(y, ymin + (.5+idx)*t);
    y = Math.min(y, ymax - (points.size()-idx-.5)*t);
    return y;
  }

  void setHeights(double t, double ymin, double ymax) {
//    System.out.println("limits: "+ymin+" "+ymax+" ; "+t);
    if (points.isEmpty()) return;

//    System.out.println("t: "+t);
    if (Holder.INSTANCE.dynamicClustering) {
      AAMap<Integer,ClusterInfo> clusters = clustering.floorEntry(t).getValue();
//      System.out.println(clusters);
      for(ClusterInfo c : clusters) {
        double s;
        int m=c.med, m2=c.med2;
        if (m2<0) s = points.get(m).y - t*(1+m);
        else s = .5*(points.get(m).y + points.get(m2).y) - .5*t*(2+m+m2);
        for(int i : c.points) {
          points.get(i).boundaryY = limitHeight(i, s + t*(1+i), ymin, ymax, t);
        }
      }
      return;
    }
//    System.out.println(medianStarting);
    if (points.size()%2==0) {
      int m = getMedianElement(t);
      int m2 = lowerStarting.floorEntry(t).getValue();
      double s = .5*(points.get(m).y + points.get(m2).y) - .5*t*(2+m+m2);
      for(int i=0; i<points.size(); ++i) {
        points.get(i).boundaryY = limitHeight(i, s + t*(1+i), ymin, ymax, t);
      }
    } else {
      int m = getMedianElement(t);
      double my = points.get(m).y;
      for(int i=0; i<points.size(); ++i) {
        points.get(i).boundaryY = limitHeight(i, my + (i-m)*t, ymin, ymax, t);
      }
    }
  }

  void generate(Set<Label> labels, boolean clustering) {
    points = new ArrayList<Label>(labels);
    if (clustering) {
      genDynamicClustering();
      return;
    }
    if (points.isEmpty()) {
      medianStarting = new TreeMap<Double,Integer>();
      heights = new TreeMap<Double,Double>();
      complexity = 0;
      return;
    }

    int mid = labels.size()/2;
    medianStarting = genKLine(mid);

    if (labels.size()%2==0) {
      int c0 = complexity;
      lowerStarting = genKLine(mid-1);
      complexity += c0;

      heights = new TreeMap<Double,Double>();
      TreeMap<Double,Integer> m1=medianStarting, m2=lowerStarting;
      Iterator<Map.Entry<Double,Integer>> i1 = m1.entrySet().iterator();
      Iterator<Map.Entry<Double,Integer>> i2 = m2.entrySet().iterator();
      Map.Entry<Double,Integer> e1 = i1.next();
      Map.Entry<Double,Integer> e2 = i2.next();
      int cur1=e1.getValue(), cur2=e2.getValue();
      double x = e1.getKey();
      e1 = i1.hasNext() ? i1.next() : null;
      e2 = i2.hasNext() ? i2.next() : null;
      x = 0;
      heights.put(x, .5*(getHeight(cur1,x) + getHeight(cur2,x)));
//      heights.put(x, getHeight(cur2,x));
      while(e1!=null || e2!=null) {
        if (e2==null || (e1!=null && e1.getKey() < e2.getKey())) {
          cur1 = e1.getValue();
          x = e1.getKey();
          e1 = i1.hasNext() ? i1.next() : null;
        } else {
          cur2 = e2.getValue();
          x = e2.getKey();
          e2 = i2.hasNext() ? i2.next() : null;
        }
//        heights.put(x, getHeight(cur2,x));
        heights.put(x, .5*(getHeight(cur1,x) + getHeight(cur2,x)));
      }
      x=5000;
//      heights.put(x, getHeight(cur2,x));
      heights.put(x, .5*(getHeight(cur1,x) + getHeight(cur2,x)));
    }
  }

  private TreeMap<Double,Integer> genKLine(int K) {
    TreeMap<Double,Integer> medians = new TreeMap<Double,Integer>();
    if (points.isEmpty()) return medians;
    Envelope upE=new Envelope(true), downE=new Envelope(false);
    for(int i=0; i<K; ++i) {
      downE.add(i);
    }
    for(int i=K+1; i<points.size(); ++i) {
      upE.add(i);
    }

    int cnum=K;
    medians.put(-5e3, cnum);
    complexity = 1;
    while(true) {
      int unum = upE.getIntersecting(cnum);
      int dnum = downE.getIntersecting(cnum);
//      System.out.printf("asd %d %d %d\n", cnum, dnum, unum);
      if (unum<0 && dnum<0) break;
      boolean up;
      if (unum<0) up=false;
      else if (dnum<0) up=true;
      else up = hitT(cnum,unum) < hitT(cnum,dnum);

      int num = up ? unum : dnum;
      Line2 ln = getLine(num);
      Envelope env = up ? upE : downE;
      env.remove(num);
      Line2 cline = getLine(cnum);
      env.add(cnum);

      double t = hitT(cnum,num);
      cnum = num;
      medians.put(t, num);

//      System.out.println("up: "+up);
      ++complexity;
    }

//    System.out.println(points);
//    System.out.println(medianStarting);
    return medians;
  }

  private Line2 getLine(int n) {
    double y = points.get(n).y;
    return new Line2(0,y,1,y-1-n);
  }
  private double hitT(int a, int b) {
    return (points.get(b).y - points.get(a).y) / (b - a);
  }

  class Envelope {
    // TODO: faster implementation

    boolean up;
    List<Integer> lines = new ArrayList<Integer>();

    Envelope(boolean up) {
      this.up = up;
    }
    void add(int n) {
      lines.add(n);
    }
    void remove(int n) {
      lines.remove(lines.indexOf(n));
    }
    int getIntersecting(int n) {
      double yn = points.get(n).y;
      double bt=1e100;
      int res=-1;
      for(int i=0; i<lines.size(); ++i) {
        int m = lines.get(i);
        if ((up && m<n) || (!up && m>n))
          continue;
        double ym = points.get(m).y;
        double t = (yn-ym)/(n-m);
        if (t<bt) {
          bt=t;
          res=m;
        }
      }
//      if (res<0) System.out.printf("No intersections for %d from %s (%b)\n", n, lines, up);
//      else System.out.printf("Intersection coming for %d: %d (%s) (%b)\n", n, res, lines, up);
      return res;
    }
    public String toString() {
      return "(Env "+lines+")";
    }
  }
  private static Envelope mergeEnvelopes(Envelope a, Envelope b) {
    if (a.lines.size() > b.lines.size()) {
      b.lines.addAll(a.lines);
      return b;
    }
    a.lines.addAll(b.lines);
    return a;
  }

  class Event implements Comparable<Event> {
    /** Is merge event. */
    boolean merge;
    /** The middle lines of an even-sizes cluster cross. */
    boolean midCross;
    /** Crossing with upper envelope. */
    boolean upper;
    /** Line which hits the median line. */
    int line;
    /** Cluster where crossing or merge occurs.
     * For merge event this is the lower of the clusters. */
    Cluster cluster;
    /** Upper cluster in a merge event. */
    Cluster cluster2;
    /** When the event occurs. */
    double t;

    Event(boolean merge, boolean midCross, boolean upper, int line, Cluster c1, Cluster c2, double t) {
      this.merge=merge;
      this.upper=upper;
      this.line=line;
      this.cluster=c1;
      this.cluster2=c2;
      this.t=t;
      this.midCross = midCross;
    }
    public int compareTo(Event e) {
      if (t!=e.t) return Double.compare(t,e.t);
      if (merge!=e.merge) return merge ? 1 : -1;
      if (midCross!=e.midCross) return midCross ? 1 : -1;
      if (upper!=e.upper) return upper ? 1 : -1;
      if (line!=e.line) return line-e.line;
      return cluster.compareTo(e.cluster);
    }
  }

  KPQ mergeKPQs(KPQ a, KPQ b, double t) {
    if (a.size > b.size) {
      KPQ k = a;
      a=b;
      b=k;
    }
    a.addAll(b, t);
    return a;
  }


  class Cluster implements Comparable<Cluster> {
//    Envelope upE=new Envelope(true), downE=new Envelope(false);
    KPQ upQ=new KPQ(), downQ=new KPQ();
    int med, med2=-1;

    double curT=0, curY;

    Event crossMid, crossUp, crossDown;
    Event mergeUp, mergeDown;

    Cluster(int m) {
      med = m;
      downQ.add(m, -points.get(m).y, 0);
    }
    public int compareTo(Cluster c) {
      if (med!=c.med) return med - c.med;
      return med2 - c.med2;
    }

    void setTakeOvers(double t) {
      setCross(false,t);
      setCross(true,t);
    }
    void setCross(boolean up, double t) {
      if (up) {
        if (crossUp!=null) events.remove(crossUp);
        crossUp = null;
      } else {
        if (crossDown!=null) events.remove(crossDown);
        crossDown = null;
      }

      KPQ k = up ? upQ : downQ;
      double tt = k.next;
//      System.out.println("Setting takeover event @ "+this+" "+t+" "+up+" "+tt+" ; "+k);
      if (tt>1e50) return;
      Event e = new Event(false, false, up, -1, this, null, tt);
      /*
      int m = up && med2>=0 ? med2 : med;
      double h = getHeight(m, t);
      int n = (up ? upE : downE).getIntersecting(m);
      if (n<0)
        return;
      double tt = getLine(m).intersection(getLine(n)).x;
      Event e = new Event(false, false, up, n, this, null, tt);
      */
      if (up)
        crossUp = e;
      else
        crossDown = e;
      events.add(e);

//      System.out.printf("added cross from %f to %f @ %s\n", t, tt, this);
    }
    void setMidCross(double t) {
      if (crossMid!=null) events.remove(crossMid);
      crossMid = null;
      double tt;
      if (med2<0) {
        tt = hitT(med, -upQ.minSlope);
      } else {
        tt = hitT(med, med2);
      }
      // FIXME: robustness
      if (tt<t)
        return;
      crossMid = new Event(false, true, false, -1, this, null, tt);
      events.add(crossMid);
    }
    void fixMergeUp() {
      if (mergeUp==null)
        return;
      Cluster c = mergeUp.cluster2;
      events.remove(mergeUp);

      double t = getMedian().intersection(c.getMedian()).x;
//      System.out.println("merge will happen at "+t+" by "+getMedian()+" and "+c.getMedian());
      Event e = new Event(true, false, false, -1, this, c, t);
      mergeUp = e;
      c.mergeDown = e;
      events.add(e);
    }
    void fixMerges() {
      fixMergeUp();
      if (mergeDown!=null)
        mergeDown.cluster.fixMergeUp();
    }
    Line2 getMedian() {
      if (med2>=0) {
        double y = .5 * (points.get(med).y + points.get(med2).y);
        double z = .5 * (2 + med + med2);
        return new Line2(0,y,1,y-z);
      }
      return getLine(med);
    }
    void mergeWithUpper(Cluster c, double t) {
      upQ = mergeKPQs(upQ, c.upQ, t);
      downQ = mergeKPQs(downQ, c.downQ, t);
      /*
      upE = mergeEnvelopes(upE, c.upE);
      downE = mergeEnvelopes(downE, c.downE);
      */
      int med1a = med, med2a = med2;
      int med1b = c.med, med2b = c.med2;
      if (med2<0) {
        if (c.med2<0) {
          med = med1b;
          med2 = med1a;

          downQ.remove(med1a, t);
          upQ.add(-med1a, points.get(med1a).y, t);
//          System.out.println("Removed "+med1a+" from downQ; "+downQ.minSlope);
        } else {
          // median already ok
        }
      } else {
        if (c.med2<0) {
          med2 = -1;
          med = c.med;
        } else {
          med = getHeight(med, t) < getHeight(c.med, t) ? c.med : med;
          med2 = getHeight(med2, t) < getHeight(c.med2, t) ? med2 : c.med2;
        }
      }
      /*
      if (med1a>=0 && med1a!=med && med1a!=med2) downE.add(med1a);
      if (med1b>=0 && med1b!=med && med1b!=med2) downE.add(med1b);
      if (med2a>=0 && med2a!=med2) upE.add(med2a);
      if (med2b>=0 && med2b!=med2) upE.add(med2b);
      */

      med = downQ.minSlope;
      if (med2>=0) med2 = -upQ.minSlope;

//      System.out.println("New medians: "+med+" "+med2);

      if (c.crossUp!=null) events.remove(c.crossUp);
      if (c.crossDown!=null) events.remove(c.crossDown);
      if (c.crossMid!=null) events.remove(c.crossMid);

//      System.out.println("Merge result: "+upQ+" "+downQ);

      mergeUp = c.mergeUp;
      fixMerges();
      setTakeOvers(t);
      setMidCross(t);
    }
    public String toString() {
//      return "{Cluster "+med+","+med2+" : "+downE+","+upE+"}";
      return "{Cluster "+med+","+med2+" : "+downQ+","+upQ+"}";
    }
  }
  double clusterHeight(Cluster c, double t) {
    if (c.med2<0) return getHeight(c.med,t);
    return .5 * (getHeight(c.med,t) + getHeight(c.med2,t));
  }

  TreeSet<Event> events;
  void genDynamicClustering() {
//    System.out.println("\n\n");
    clustering = new TreeMap<Double,AAMap<Integer,ClusterInfo>>();
    events = new TreeSet<Event>();
    clusterLines = new ArrayList<Line2D.Double>();
    TreeSet<Cluster> clusters = new TreeSet<Cluster>();
    Cluster prevC=null;
    AAMap<Integer,ClusterInfo> infos = new AAMap<Integer,ClusterInfo>();

    if (points.isEmpty()) return;

    for(int i=0; i<points.size(); ++i) {
      Label l = points.get(i);
      Cluster c = new Cluster(i);
      clusters.add(c);
      c.curY = getHeight(i, c.curT);

//      ArrayList<Integer> lst = new ArrayList<Integer>();
//      lst.add(i);
      AASet<Integer> lst = new AASet<Integer>(i);
      infos = infos.put(i, new ClusterInfo(lst, i, -1));

      if (prevC!=null) {
        double t = getLine(i-1).intersection(getLine(i)).x;
        Event merge = new Event(true, false, false, -1, prevC, c, t);
        events.add(merge);
        prevC.mergeUp = merge;
        c.mergeDown = merge;
      }
      prevC=c;
    }

    clustering.put(-5e3, infos);

    complexity = points.size();
    double prevT=0;
    while(!events.isEmpty()) {
      ++complexity;
      Event evt = events.pollFirst();
      Cluster c = evt.cluster;
      int oldMed = c.med;
      double ct = c.curT;

//      System.out.println("current t "+evt.t);
      if (evt.merge) {
        // Clusters merge.
//        System.out.printf("Merging clusters %f: %s %s\n", evt.t, c, evt.cluster2);

        clusters.remove(evt.cluster2);
        c.mergeWithUpper(evt.cluster2, evt.t+1e-6);
      } else if (evt.midCross) {
        // Lines below and above median are swapped.
//        System.out.println("middle takeover @ "+c);

        double t = evt.t+1e-6;
        int med2 = c.med2<0 ? -c.upQ.minSlope : c.med2;
        c.downQ.remove(c.med, t);
        c.upQ.remove(-med2, t);
        c.upQ.add(-c.med, points.get(c.med).y, t);
        c.downQ.add(med2, -points.get(med2).y, t);

        int m = c.med;
        c.med = med2;
        if (c.med2>=0)
          c.med2 = m;

        c.crossMid = null;
        c.setTakeOvers(t);
        if (c.med2<0) {
          c.setMidCross(t);
          c.fixMerges();
        }
      } else {
        // Median line of a cluster changes.
//        System.out.println("takeover event @ "+c+" ; "+evt.upper);

        /*
        Envelope e = evt.upper ? c.upE : c.downE;
        boolean upmed = evt.upper && c.med2>=0;
        int m = upmed ? c.med2 : c.med;
        double t = evt.t;
        double h = getHeight(m, t);
//        int num = e.getIntersecting(new Point2(t,h), new Vec2(1, -1-m));
        int num = evt.line;

//        System.out.println("asd num: "+num);
        e.remove(num);
        e.add(m);
        */
        KPQ k = evt.upper ? c.upQ : c.downQ;
        boolean upmed = evt.upper && c.med2>=0;
        if (!k.advance()) {
//          System.out.println("dummy takeover");
          c.setCross(evt.upper, evt.t);
          continue;
        }
        int num = evt.upper ? -k.minSlope : k.minSlope;

        if (!evt.upper || c.med2>=0) {
          if (upmed)
            c.med2 = num;
          else
            c.med = num;
        }

        double tt = evt.t;
        c.setTakeOvers(tt);
        c.setMidCross(tt);
        c.fixMerges();
      }
//      System.out.println(c);

      prevT = evt.t;
      c.curT = evt.t;

//      infos = infos.remove(oldMed);

//      AASet<ClusterInfo> oldI = infos;
//      infos = new AASet<ClusterInfo>(oldI);
      if (evt.merge) {
        /*
        removeCluster(infos, oldMed);
        if (evt.cluster2!=null) removeCluster(infos, evt.cluster2.med);
        infos.add(makeClusterInfo(evt.cluster));
        */
//        System.out.println("getting "+oldMed+" from "+infos);
        ClusterInfo old = infos.get(oldMed);
        ClusterInfo old2 = infos.get(evt.cluster2.med);
//        System.out.println("merging "+old+" and "+old2+"  @ "+infos);
        AASet<Integer> aa = mergeSets(old.points, old2.points);
        ClusterInfo cc = new ClusterInfo(aa, c.med, c.med2);
        infos = infos.remove(oldMed).remove(evt.cluster2.med).put(c.med, cc);
      } else {
        ClusterInfo old = infos.get(oldMed);
        infos = infos.remove(oldMed).put(c.med, new ClusterInfo(old.points, c.med, c.med2));
        /*
        int cnum = 0;
        while(infos.get(cnum).med != oldMed) {
          ++cnum;
        }
        infos.set(cnum, new ClusterInfo(infos.get(cnum).points, c.med, c.med2));
        */
      }
//      System.out.println("result at "+evt.t+": "+infos);
      clustering.put(evt.t, infos);

      double h = clusterHeight(c, evt.t);
      clusterLines.add(new Line2D.Double(ct,c.curY,evt.t,h));
      if (evt.cluster2 != null) {
        clusterLines.add(new Line2D.Double(evt.cluster2.curT,evt.cluster2.curY,evt.t,h));
      }
      c.curY = h;
      c.curT = evt.t;
    }
    Cluster cl = clusters.first();
    double xx = 5e3;
//    while(Math.abs(clusterHeight(cl,xx))>1500) xx*=.5;
    clusterLines.add(new Line2D.Double(prevT,
      clusterHeight(cl,prevT),
      xx,
      clusterHeight(cl,xx)
    ));
  }

  /*
  private static void removeCluster(ArrayList<ClusterInfo> lst, int med) {
      int mnum = 0;
      while(lst.get(mnum).med != med) {
        ++mnum;
      }
      lst.remove(mnum);
  }
  private ClusterInfo makeClusterInfo(Cluster c) {
//    System.out.println("making cl info "+c);
    ArrayList<Integer> nums = new ArrayList<Integer>();
//    nums.addAll(c.downE.lines);
    c.downQ.addTo(nums, -1);
//    nums.add(c.med);
//    if (c.med2>=0) nums.add(c.med2);
//    nums.addAll(c.upE.lines);
    c.upQ.addTo(nums, 1);
//    System.out.println("nums: "+nums);
    return new ClusterInfo(nums, c.med, c.med2);
  }
  */
  private static AASet<Integer> mergeSets(AASet<Integer> a, AASet<Integer> b) {
    if (b.getLevel() < a.getLevel()) {
      AASet<Integer> t=a;
      a=b;
      b=t;
    }
    for(int i : a)
      b = b.add(i);
    return b;
  }
}

class Vec2 extends Point2D.Double {
  Vec2(double x, double y) {
    super(x,y);
  }

  Vec2 multiply(double a) {
    return new Vec2(a*x, a*y);
  }
  double cross(Vec2 v) {
    return x*v.y - y*v.x;
  }
  double dot(Vec2 v) {
    return x*v.x + y*v.y;
  }
}
class Point2 extends Point2D.Double {
  Point2(double x, double y) {
    super(x,y);
  }

  Vec2 subtract(Point2 p) {
    return new Vec2(x-p.x, y-p.y);
  }
  Vec2 vectorTo(Point2 p) {
    return new Vec2(p.x-x, p.y-y);
  }
  Point2 add(Vec2 v) {
    return new Point2(x+v.x, y+v.y);
  }
  public String toString() {
    return "("+x+","+y+")";
  }
}

class Line2 extends Line2D.Double {
  Line2(Point2 a, Point2 b) {
    super(a.x,a.y,b.x,b.y);
  }
  Line2(Point2 a, Vec2 v) {
    super(a.x,a.y,a.x+v.x,a.y+v.y);
  }
  Line2(double x1, double y1, double x2, double y2) {
    super(x1,y1,x2,y2);
  }
  Point2 start() {
    return new Point2(x1,y1);
  }
  Point2 end() {
    return new Point2(x2,y2);
  }
  Vec2 direction() {
    return new Vec2(x2-x1, y2-y1);
  }
  boolean crossesRay(Point2 p, Vec2 d) {
    Vec2 v = direction();
    Point2 st = start();
    Point2 ed = end();
    return v.cross(st.vectorTo(p)) * v.cross(d) < 0;
  }

  Point2 intersection(Line2 l) {
    //a + (b-a)*fabs(cross(c-a,d-c)/cross(b-a,d-c))
    Vec2 d = l.direction();
    double a = l.start().subtract(start()).cross(d);
    double b = direction().cross(d);
    double z = Math.abs(a/b);
    return start().add(direction().multiply(z));
  }

  public String toString() {
    return "["+start() + " - "+end()+"]";
  }
}
