package gui;

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.event.MouseEvent;
import javax.swing.*;
import java.awt.*;
import java.awt.image.*;

import data.BL;
import data.Holder;
import data.Label;

public class LabelCanvas extends Canvas {
  private static final long serialVersionUID = -2061991344737587501L;

  ZoomAndPanListener zoomAndPanListener;
  private boolean init = true;
  private Rectangle2D labelArea;

  public boolean internalLabels = false;

  private Image offscreen;
  Image background = null;

  LabelingPanel panel;

  public LabelCanvas(Rectangle2D labelArea, int width, int height, LabelingPanel panel) {
    this.panel = panel;
    setBackground(Color.WHITE);
    // some space for labels (label box size defined in Holder, rest for the area
    System.out.println("Label canvas w "+width);
    setPreferredSize(new Dimension(width, height));
    this.labelArea = labelArea;

    this.zoomAndPanListener = new ZoomAndPanListener(new ZoomAndPanAction() {
      public void mousePressed(MouseEvent e) {
        if (e.getButton()==MouseEvent.BUTTON3) {
          String labelText = (String) JOptionPane.showInputDialog(new JFrame(),
              "Enter text for new label:\n", "Label Dialog",
              JOptionPane.PLAIN_MESSAGE);
          
          if (labelText != null) {
            try {
              Point2D.Float pf = zoomAndPanListener.transformPoint(e.getPoint());
              Label label = new Label(labelText, pf.getX(), pf.getY());
              Holder.INSTANCE.addLabel(label);
              Holder.INSTANCE.window.repaint();
            } catch (NoninvertibleTransformException e1) {
              // TODO Auto-generated catch block
              e1.printStackTrace();
            }
          }
        }
      }
      public void repaint() {
        Holder.INSTANCE.window.repaint();
      }
      public void setZoom(double level) {
        Holder.INSTANCE.setZoomLevel(level);
      }
    }, 1.05);

    this.addMouseListener(zoomAndPanListener);
    this.addMouseMotionListener(zoomAndPanListener);
    this.addMouseWheelListener(zoomAndPanListener);
  }

/*
  public Dimension getPreferredSize() {
    return new Dimension(WindowInterface.LABEL_AREA_WIDTH, WindowInterface.PAINT_AREA_HEIGHT);
  }*/

  @Override
  public void paint(Graphics graphics) {
    Graphics2D g = (Graphics2D) graphics;
    if (!g.getClipBounds().getSize().equals(getSize())) {
      System.out.println("repainting... "+getBounds()+" "+g.getClipBounds());
      repaint(10);
      return;
    }

    if (offscreen==null || offscreen.getWidth(null)!=getWidth() || offscreen.getHeight(null)!=getHeight()) {
      offscreen = createImage(getWidth(), getHeight());
    }
    Graphics2D oldg = g;
    g = (Graphics2D)offscreen.getGraphics();
    g.clearRect(0,0,getWidth(),getHeight());
    g.setClip(oldg.getClipBounds());
    g.setTransform(oldg.getTransform());
    g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

    paintCanvas(g);

    g.dispose();
    oldg.drawImage(offscreen, 0, 0, null);
  }

  void paintCanvas(Graphics2D g) {
    double labelW = WindowInterface.LABEL_AREA_WIDTH;
    labelArea = new Rectangle2D.Double(getWidth()-labelW, 0, labelW, getHeight());

    FontMetrics fm = g.getFontMetrics();

    if (init) {
      // Initialize the viewport by moving the origin to the center of the
      // window and invert the y-axis to point upwards.
      init = false;
      Dimension d = getSize();
//      g.translate(labelArea.getMaxX() - 1, d.height);
      g.translate(0, d.height);
      g.scale(1, -1);
      // Save the viewport to be updated by the ZoomAndPanListener
//      zoomAndPanListener.setCoordTransform(g.getTransform());

      Holder.INSTANCE.setZoomLevel(zoomAndPanListener.getZoomLevel());
    } else {
      // Restore the viewport after it was updated by the ZoomAndPanListener
      limitView();
      g.setTransform(zoomAndPanListener.getCoordTransform());
    }


    if (background != null) {
//      g.drawImage(background, 0, 0, null);
      AffineTransform at = AffineTransform.getScaleInstance(1,-1);
      at.translate(0, -background.getHeight(null));
      g.drawImage(background, at, null);
    }

    // Draw axes
    g.setColor(Color.RED);
    g.drawLine(-15000, 0, 15000, 0);
    g.drawLine(0, -15000, 0, 15000);

//    Rectangle bounds = g.getClipBounds();
    Rectangle bounds = getBounds();
    bounds.width -= labelW;
    if (Holder.INSTANCE.checkVisible(g.getTransform(), bounds, fm, internalLabels)) {
      panel.labelsChanged();
    }

    g.setColor(Color.BLACK);
    // Create an "upside-down" font to correct for the inverted y-axis
    Font font = g.getFont();
    AffineTransform affineTransform = new AffineTransform();
    affineTransform.scale(1, -1);
    g.setFont(font.deriveFont(affineTransform));
    // Draw the points and their coordinates

    /*
    // get the clipping area for matching which points are visible
    // and which not -- should be done elsewhere, but oh well..
    Rectangle2D area = g.getClipBounds();
    double scaledSize = labelArea.getWidth() * zoomAndPanListener
        .getCurrentZoomMultiplicationFactor();
    area.setRect(area.getX() + scaledSize, area.getY(), area.getWidth()
        - scaledSize, area.getHeight());
        */

    for (Label label : Holder.INSTANCE.labels) {
      if (internalLabels)
        g.drawString(label.text, (int) label.x, (int) label.y + 3);
//      int s = 4;
//      g.fillOval((int) label.x - s/2, (int) label.y - s/2, s, s);
    }

    Rectangle oldBounds = bounds;
    bounds = g.getClipBounds();
//    System.out.println("drawing with clip bounds: "+bounds);

    AffineTransform at = zoomAndPanListener.getCoordTransform();
    affineTransform = new AffineTransform();
    g.setTransform(affineTransform);
    font = g.getFont();
    g.setFont(font.deriveFont(affineTransform));
    // Draw background for labels
//    g.setColor(Color.WHITE);
    g.setColor(new Color(200,200,200));
    g.fill(labelArea);
    g.setColor(Color.BLACK);
       
    int labelAreaWidth = (int) labelArea.getWidth();
    Holder.INSTANCE.setHeights(
      BL.getRectangleSize(font.getSize()),
      zoomAndPanListener.getCurrentZoomMultiplicationFactor(),
      bounds.y,
      bounds.y + bounds.height
    );
    BL.generateLabels(at, font.getSize(), labelAreaWidth, labelArea.getX());
    
    int x0 = getWidth() - WindowInterface.LABEL_AREA_WIDTH + 2;
    int idx = 1;
    g.setTransform(affineTransform);
    for(Label l : Holder.INSTANCE.allLabels) {
      Point2D dbl = at.transform(l, null);
      if (!oldBounds.contains(dbl)) continue;
      int s=4;
      g.fillOval((int)dbl.getX()-s/2, (int)dbl.getY()-s/2, s, s);
    }
    for(Label l: Holder.INSTANCE.labels) {
      l.paintLabelBox(g, font.getSize(), idx);
      idx++;
      
      int midPoint = (int) l.labelRectangle.getCenterY();
//        int x0 = WindowInterface.PAINT_AREA_WIDTH+2;
      g.fillOval(x0, midPoint+1, 2, 2);

      // draw leads
      Point2D dbl = new Point2D.Double(l.x, l.y);
      dbl = at.transform(l, dbl);
//      System.out.println(dbl+" ; "+(getWidth()-labelAreaWidth+2));
//        System.out.printf("new Label(\"%s\", %f, %f),\n", l.text, l.x, l.y);
      if (dbl.getX() >= x0) continue;
      g.setColor(Color.GRAY);
      int y0 = midPoint+2;
      g.drawLine(x0, y0, (int)dbl.getX(), y0);
      if (y0 != (int)dbl.getY())
        g.drawLine((int)dbl.getX(), y0, (int) dbl.getX(), (int) dbl.getY());
      g.setColor(Color.BLACK);
    }
  }

  public void update(Graphics g) {
    paint(g);
  }

  public void resetView(double w, double h) {
    System.out.println("view: "+w+" "+h);
    double ww = getWidth() - WindowInterface.LABEL_AREA_WIDTH+1;
    zoomAndPanListener.resetView(.5*w, .5*h, .95*Math.min(ww/w, getHeight()/h), getHeight());
  }

  void limitView() {
    if (Holder.INSTANCE.allLabels.isEmpty()) return;
    AffineTransform at = zoomAndPanListener.getCoordTransform();
    double xmin=1e50, xmax=-1e50, ymin=1e50, ymax=-1e50;
    for(Label l : Holder.INSTANCE.allLabels) {
      Point2D p = at.transform(l, null);
      double x = p.getX();
      double y = p.getY();
      xmin = Math.min(xmin, x);
      ymin = Math.min(ymin, y);
      xmax = Math.max(xmax, x);
      ymax = Math.max(ymax, y);
    }
    double dx=0, dy=0;
    double w = getWidth()-WindowInterface.LABEL_AREA_WIDTH, h = getHeight();
    if (xmin > w) {
      dx = w-xmin-1e-5;
    } else if (xmax<0) {
      dx = 1e-5-xmax;
    }
    if (ymin > h) {
      dy = h-ymin-1e-5;
    } else if (ymax<0) {
      dy = 1e-5-ymax;
    }
    if (dx!=0 || dy!=0) {
//      System.out.println("Breaking view: "+dx+" "+dy);
      at.translate(dx / at.getScaleX(), dy / at.getScaleY());
      zoomAndPanListener.setCoordTransform(at);
    }
  }
}
