package gui;

import java.awt.BorderLayout;
import java.awt.geom.Rectangle2D;

import javax.swing.*;
import javax.swing.text.MaskFormatter;
import javax.swing.text.*;

import java.awt.event.*;
import java.awt.image.BufferStrategy;
import java.awt.*;
import javax.swing.event.*;

import javax.swing.filechooser.FileNameExtensionFilter;
import java.io.*;
import javax.imageio.ImageIO;

import data.Holder;

public class LabelingPanel extends JPanel implements WindowInterface, ActionListener {
  
  final Rectangle2D labelArea = new Rectangle2D.Double(PAINT_AREA_WIDTH, 0, LABEL_AREA_WIDTH, PAINT_AREA_HEIGHT);
  final LabelCanvas labelCanvas = new LabelCanvas(labelArea, LABEL_AREA_WIDTH+PAINT_AREA_WIDTH, PAINT_AREA_HEIGHT, this);
  final VectorCanvas vectorCanvas = new VectorCanvas(PAINT_AREA_WIDTH, PAINT_AREA_HEIGHT);
  
  /*
  public static final JSlider zoomSlider = new JSlider(ZoomAndPanListener.MIN_ZOOM_LEVEL, ZoomAndPanListener.MAX_ZOOM_LEVEL, 0);
  */
  
  public JMenuBar menuBar;

  PriorityList priorities = new PriorityList();
  
  public LabelingPanel() {
    super();

    createAndShowGUI();
    Holder.INSTANCE.window = this;
  }

  private JComponent makeChoises() {
    JPanel choises = new JPanel();
    choises.setLayout(new BoxLayout(choises, BoxLayout.X_AXIS));
    choises.add(makeButtons());
//    choises.add(new JSeparator(JSeparator.VERTICAL));
    choises.add(new PriorityChooser());
//    choises.add(new JSeparator(JSeparator.VERTICAL));
    choises.add(new OptionsPanel(this));
    return choises;
  }
  private JComponent makeButtons() {
    JPanel buttons = new JPanel();
    buttons.setLayout(new BoxLayout(buttons, BoxLayout.Y_AXIS));

    JButton clearButton = new JButton("Clear");
    clearButton.addActionListener(Holder.INSTANCE.listener);
    buttons.add(clearButton);
    buttons.add(new RandomGenerator());

    buttons.setMaximumSize(new Dimension(240,100));

    return buttons;
  }
  public class RandomGenerator extends JPanel {
    JFormattedTextField text;
    JButton button = new JButton("Generate random instance");
    RandomGenerator() {

      setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
      text = new JFormattedTextField(new NumberFormatter());
      text.setMaximumSize(new Dimension(200, 20));

//      JPanel p = new JPanel();
//      p.setLayout(new BoxLayout(p, BoxLayout.X_AXIS));
//      p.add(new JLabel("Random site count: "));
//      p.add(text);
//      add(p);

      add(new JLabel("Random site count:"));
      add(text);
      add(button);

      button.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
//          Holder.INSTANCE.generateRandom(Integer.parseInt(text.getText()));
          Long val = (Long)text.getValue();
          if (val!=null) {
            Dimension d = Holder.INSTANCE.generateRandom((int)val.longValue());
            labelCanvas.resetView(d.width, d.height);
            LabelingPanel.this.repaint();
          }
        }
      });

//      setBorder(BorderFactory.createLineBorder(Color.black));
    }
  }
  public class PriorityChooser extends JPanel implements CaretListener {
    JFormattedTextField text = new JFormattedTextField(new NumberFormatter());

    PriorityChooser() {
      text.setValue(new Long(Holder.INSTANCE.maxVisible));
      setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
      add(new JLabel("Maximum visible labels"));
      add(text);
      text.setMaximumSize(new Dimension(200, 20));
      text.addCaretListener(this);
    }
    public void caretUpdate(CaretEvent e) {
      int cnt = 0;
      try {
        text.commitEdit();
        Long val = (Long)text.getValue();
        if (val!=null) cnt = (int)val.longValue();
      } catch(java.text.ParseException ee) {
      }
      System.out.println("yeah "+cnt);
      Holder.INSTANCE.maxVisible = cnt;
      LabelingPanel.this.repaint();
    }
  }

  private boolean firstPaint = true;
  @Override
  public void repaint() {
    if (labelCanvas==null) return;
    if (firstPaint) {
      double x=0, y=0;
      for(data.Label l : Holder.INSTANCE.allLabels) {
        x = Math.max(x, l.x);
        y = Math.max(y, l.y);
      }
      labelCanvas.resetView(x, y);
      firstPaint = false;
    }
    super.repaint();
    labelCanvas.repaint();
    vectorCanvas.repaint();

/*
    BufferStrategy bf = getBufferStrategy();
    Graphics g = null;
    try {
      g = bf.getDrawGraphics();
      labelCanvas.paint(g);
      vectorCanvas.paint(g);
    } finally {
      g.dispose();
    }
    bf.show();
    Toolkit.getDefaultToolkit().sync();
    */
  }

  /**
   * Create a menu from the description and add it to the menu bar.
   *
   * @param name User-visible menu name
   * @param desc Accessible description for menu
   * @param mnemonic Menu mnemonic
   * @param items Menu items in format { name, command, desc, mnemonic }
   **/
  private void createMenu(String name, String desc,
      String mnemonic, Object[][] items)
  {
    JMenu menu = new JMenu(name);
    menu.setMnemonic(mnemonic.charAt(0));
    menu.getAccessibleContext().setAccessibleDescription(desc);

    for (Object[] i : items) {
      String
        iName = (String) i[0],
              iCommand = (String) i[1],
              iDesc = (String) i[2];

      JMenuItem item = new JMenuItem(iName);
      item.getAccessibleContext().setAccessibleDescription(iDesc);
      item.setActionCommand(iCommand);

      if (i[3] != null)
        item.setMnemonic(((String) i[3]).charAt(0));

      item.addActionListener(this);
      menu.add(item);
    }

    menuBar.add(menu);
  }

  /**
   * Constructs a main window and makes it visual.
   **/
  private void createAndShowGUI() {
    menuBar = new JMenuBar();
    Object[][] fileItems = {
      { "Load background image", "background", "Load background image from file", "b" },
      { "Export merge tree view as SVG", "medians", "Export merge tree as SVG", "m" },
      { "Export label view as SVG", "map", "Export map image as SVG", "l" }
    };
    createMenu("File", "Menu", "f", fileItems);

    // create canvases
    setLayout(new BorderLayout());

    JSplitPane split = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, labelCanvas, vectorCanvas);
    split.setContinuousLayout(true);
    add(split, BorderLayout.CENTER);
    Dimension small = new Dimension(100,100);
    labelCanvas.setMinimumSize(small);
    vectorCanvas.setMinimumSize(small);
    split.setDividerLocation(LABEL_AREA_WIDTH+PAINT_AREA_WIDTH);

    add(makeChoises(), BorderLayout.SOUTH);
    add(priorities, BorderLayout.EAST);
    
    // set params
//    setSize(LABEL_AREA_WIDTH+PAINT_AREA_WIDTH+64, PAINT_AREA_HEIGHT+120);

//    createBufferStrategy(2);
//    canvases.setDoubleBuffered(true);
    
  }

  public void actionPerformed(ActionEvent e) {
    String c = e.getActionCommand();
    if (c.equals("medians")) {
      exportMedians();
    } else if (c.equals("background")) {
      loadBackground();
    } else if (c.equals("map")) {
      exportMap();
    }
  }

  private void exportMedians() {
    JFileChooser fc = new JFileChooser();
    fc.setFileFilter(new FileNameExtensionFilter
        ("Scalable Vector Graphics (SVG)", "svg"));
    if (fc.showSaveDialog(this) != JFileChooser.APPROVE_OPTION)
      return;

    try {
      File file = fc.getSelectedFile();
      if (!file.getName().endsWith(".svg"))
        file = new File(file.getPath() + ".svg");

      Rectangle r = vectorCanvas.getBounds();
      r.x=0;
      r.y=0;
      new SVGWriter().drawMergeTree(file, r, vectorCanvas.zoomAndPanListener.getCoordTransform());
    } catch(IOException e) {
      JOptionPane.showMessageDialog
        (null, "Writing to file failed: "+e.getMessage());
    }
  }
  private void exportMap() {
    JFileChooser fc = new JFileChooser();
    fc.setFileFilter(new FileNameExtensionFilter
        ("Scalable Vector Graphics (SVG)", "svg"));
    if (fc.showSaveDialog(this) != JFileChooser.APPROVE_OPTION)
      return;

    try {
      File file = fc.getSelectedFile();
      if (!file.getName().endsWith(".svg"))
        file = new File(file.getPath() + ".svg");

      Rectangle r = labelCanvas.getBounds();
      r.x=0;
      r.y=0;
      new SVGWriter().drawMap(file, r, labelCanvas.zoomAndPanListener.getCoordTransform());
    } catch(IOException e) {
      JOptionPane.showMessageDialog
        (null, "Writing to file failed: "+e.getMessage());
    }
  }
  private void loadBackground() {
    JFileChooser fc = new JFileChooser();
    fc.setFileFilter(new FileNameExtensionFilter
        ("Image files", "png","jpg","gif"));
    if (fc.showOpenDialog(this) != JFileChooser.APPROVE_OPTION)
      return;

    try {
      File file = fc.getSelectedFile();
      Image bg = ImageIO.read(file);
      labelCanvas.background = bg;
    } catch(IOException e) {
      JOptionPane.showMessageDialog
        (null, "Loading background failed: "+e.getMessage());
    }
  }

  public void labelsChanged() {
    priorities.generate();
    repaint();
  }
}
