package gui;

import data.*;

import java.io.*;
import java.util.List;
import java.util.HashMap;

import java.awt.Rectangle;
import java.awt.geom.*;
import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;

public class SVGWriter {

  public void drawMergeTree(File f, Rectangle bounds, AffineTransform at) throws IOException {
    System.out.println("CREATING SVG: "+bounds);
    PrintStream out = new PrintStream(f);
    out.println(svgStart(bounds));

    // draw axis
    out.println(lineNode(Util.clipLine(-5000,0,5000,0, bounds, at), "blue", 3));
    out.println(lineNode(Util.clipLine(0,-5000,0,5000, bounds, at), "blue", 3));

    // draw zoom level
    double zoomX = Holder.INSTANCE.getZoomFactor();
    out.println(lineNode(Util.clipLine(zoomX,-5000,zoomX,5000, bounds, at), "green", 3));

    // draw lines
    List<Line2D.Double> lines = Holder.INSTANCE.dynamicClustering
      ? Holder.INSTANCE.getClusterLines()
      : VectorCanvas.genMedianLine();
    for(Line2D.Double l : lines) {
      Point2D a = at.transform(new Point2D.Double(l.x1,l.y1), null);
      Point2D b = at.transform(new Point2D.Double(l.x2,l.y2), null);
      Line2D.Double ll = Util.clipLine(a.getX(), a.getY(), b.getX(), b.getY(), bounds);
      if (ll==null) continue;
      out.println(lineNode(ll, "red", 4));
    }

    int idx=1;
    for(Label l : Holder.INSTANCE.labels) {
      Line2D.Double d = Util.clipLine(l.getDualLine(idx, at), bounds);
      if (d!=null) {
        out.println(lineNode(d, "black", 1));
      }
      ++idx;
    }

    out.println("</svg>");
    out.close();
  }
  public void drawMap(File f, Rectangle bounds, AffineTransform at) throws IOException {
    PrintStream out = new PrintStream(f);
    out.println(svgStart(bounds));

    out.println(lineNode(Util.clipLine(-5000,0,5000,0, bounds, at), "red", 1));
    out.println(lineNode(Util.clipLine(0,-5000,0,5000, bounds, at), "red", 1));

    double labelW = WindowInterface.LABEL_AREA_WIDTH;
    double paintW = bounds.getWidth() - labelW;
    out.println(rectNode(paintW, 0., labelW, bounds.getHeight(), "none", "#cccccc"));

//    out.printf("<g tranform=\"translate=(%f,%f) scale=(%f,%f)\">\n", at.getTranslateX(), at.getTranslateY(), at.getScaleX(), at.getScaleY());
    for(Label l : Holder.INSTANCE.labels) {
      Point2D p = at.transform(l, null);
      Point2D q = at.transform(new Point2D.Double(l.x, l.boundaryY), null);
//      Point2D q = new Point2D.Double(p.getX(), l.boundaryY);
      double rsize = 16;
      Rectangle2D.Double r = l.labelRectangle;
      out.println(rectNode(r.x, r.y, r.width, r.height, "black", "none"));
//      out.println(rectNode(paintW, q.getY()-.5*rsize, labelW, rsize, "black", "none"));
//      out.println(textNode(l.text, "black", paintW+2, q.getY(), 12, "middle"));
      int fsize = 12;
      out.println(textNode(l.text, "black", paintW+2, r.y + fsize, fsize));
      if (Holder.INSTANCE.window.labelCanvas.internalLabels) {
        out.println(textNode(l.text, "black", p.getX(), p.getY()-1, 10*at.getScaleX()));
      }

      out.printf("<circle cx=\"%f\" cy=\"%f\" r=\"%f\"/>\n", p.getX(), p.getY(), 2.);
      String color = "grey";
      out.println(lineNode(p, q, color, 1));
      out.println(lineNode(q, new Point2D.Double(paintW, q.getY()), color, 1));

      System.out.printf("lol y of %s : %f %f\n", l, l.y, l.boundaryY);
    }

    out.println("</svg>");
    out.close();
  }

  private String textNode(String text, String colorstr, double x, double y, double s) {
    return textNode(text, colorstr, x, y, s, "alphabetic");
  }
  private String textNode(String text, String colorstr, double x, double y, double s, String base) {
    StringBuffer ret = new StringBuffer("<text x=\"");
    ret.append(x);
    ret.append("\" y=\"");
    ret.append(y);
//    ret.append("\" font-family=\"Verdana\" font-size=\"100\" font-weight=\"bold\" fill=\"");
    ret.append("\" font-family=\"Verdana\" font-size=\""+s+"\" fill=\"");
    ret.append(colorstr);
    ret.append("\" alignment-baseline=\"");
    ret.append(base);
    ret.append("\" text-anchor=\"start\" >");
    ret.append(text);
    ret.append("</text>");
    return ret.toString();
  }
  private static String lineNode(Line2D.Double l, String color, int width) {
    if (l==null) return "";
    return String.format("<line x1=\"%f\" y1=\"%f\" x2=\"%f\" y2=\"%f\" stroke=\"%s\" stroke-width=\"%d\"/>",
        l.x1, l.y1, l.x2, l.y2, color, width);
  }
  private static String lineNode(Point2D a, Point2D b, String color, int width) {
    return lineNode(new Line2D.Double(a.getX(), a.getY(), b.getX(), b.getY()), color, width);
  }
  private static String lineNode(Point2D a, Point2D b, AffineTransform at, String color, int width) {
    at.transform(a, a);
    at.transform(b, b);
    return lineNode(new Line2D.Double(a.getX(), a.getY(), b.getX(), b.getY()), color, width);
  }
  private static String rectNode(double x, double y, double w, double h, String stroke, String fill) {
    return String.format("<rect x=\"%f\" y=\"%f\" width=\"%f\" height=\"%f\" stroke=\"%s\" fill=\"%s\"/>",
        x, y, w, h, stroke, fill);
  }

  private static String svgStart(Rectangle bounds) {
    return "<?xml version=\"1.0\" standalone=\"no\"?>\n"
      + "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 20010904//EN\" \"http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd\">\n"
      + String.format("<svg width=\"%dpx\" height=\"%dpx\" viewBox=\"%d %d %d %d\" xmlns=\"http://www.w3.org/2000/svg\">\n",
          bounds.width, bounds.height, bounds.x, bounds.y, bounds.width, bounds.height);
  }
}
