package gui;

import java.awt.*;
import java.awt.geom.*;
import javax.swing.*;

public class Util {
  // Hack to avoid drawing too long lines as it breaks some swing implementations
  public static void drawLine(Graphics2D g, double x1, double y1, double x2, double y2) {
    double xs = g.getTransform().getScaleX();
    double ys = g.getTransform().getScaleY();

    Rectangle r = g.getClipBounds();
    if (r==null) {
      System.out.println("Warning: drawing line without clipping.");
      g.drawLine((int)x1,(int)y1,(int)x2,(int)y2);
    } else {
//      System.out.printf("Trying to draw line: %f %f %f %f\n",x1,y1,x2,y2);
      Line2D.Double l = clipLine(x1, y1, x2, y2, r);
      if (l!=null) {
//        System.out.printf("clip result: %f %f %f %f\n", l.x1,l.y1,l.x2,l.y2);
        g.drawLine((int)l.x1, (int)l.y1, (int)l.x2, (int)l.y2);
      }
    }
  }
  public static void drawLine(Graphics2D g, Line2D l) {
    drawLine(g, l.getX1(), l.getY1(), l.getX2(), l.getY2());
  }
  public static void drawLine(Graphics2D g, Point2D a, Point2D b) {
    drawLine(g, a.getX(), a.getY(), b.getX(), b.getY());
  }
  public static void drawLine(Graphics2D g, double x1, double y1, double x2, double y2, AffineTransform at) {
    drawLine(g, new Point2D.Double(x1,y1), new Point2D.Double(x2,y2), at);
  }
  public static void drawLine(Graphics2D g, Point2D a, Point2D b, AffineTransform at) {
    drawLine(g, at.transform(a,null), at.transform(b,null));
  }

  public static Line2D.Double clipLine(Line2D l, Rectangle r) {
    return clipLine(l.getX1(), l.getY1(), l.getX2(), l.getY2(), r);
  }
  public static Line2D.Double clipLine(double x1, double y1, double x2, double y2, Rectangle r, AffineTransform at) {
    return clipLine(
        at.transform(new Point2D.Double(x1, y1),null),
        at.transform(new Point2D.Double(x2,y2),null),
        r);
  }
  public static Line2D.Double clipLine(Point2D a, Point2D b, Rectangle r) {
    return clipLine(a.getX(), a.getY(), b.getX(), b.getY(), r);
  }

  // taken from http://en.wikipedia.org/wiki/Cohen–Sutherland
  public static Line2D.Double clipLine(double x1, double y1, double x2, double y2, Rectangle r) {
    //Outcodes for P0, P1, and whatever point lies outside the clip rectangle
    int outcodeOut, hhh = 0;
    boolean accept = false, done = false;

    //compute outcodes
    int outcode0 = r.outcode(x1, y1);
    int outcode1 = r.outcode(x2, y2);

    double xmin=r.getX(), ymin=r.getY();
    double xmax=xmin+r.getWidth(), ymax=ymin+r.getHeight();

    do{
      if ((outcode0 | outcode1) == 0 )
      {
        accept = true;
        done = true;
      }
      else if ( (outcode0 & outcode1) > 0 )
      {
        done = true;
      }

      else
      {
        //failed both tests, so calculate the line segment to clip
        //from an outside point to an intersection with clip edge
        double x = 0, y = 0;
        //At least one endpoint is outside the clip rectangle; pick it.
        outcodeOut = outcode0 != 0 ? outcode0: outcode1;
        //Now find the intersection point;
        //use formulas y = y0 + slope * (x - x0), x = x0 + (1/slope)* (y - y0)
        if ( (outcodeOut & Rectangle2D.OUT_BOTTOM) > 0 )
        {
          x = x1 + (x2 - x1) * (ymax - y1)/(y2 - y1);
          y = ymax;
        }
        else if ((outcodeOut & Rectangle2D.OUT_TOP) > 0 )
        {
          x = x1 + (x2 - x1) * (ymin - y1)/(y2 - y1);
          y = ymin;
        }
        else if ((outcodeOut & Rectangle2D.OUT_RIGHT)> 0)
        {
          y = y1 + (y2 - y1) * (xmax - x1)/(x2 - x1);
          x = xmax;
        }
        else if ((outcodeOut & Rectangle2D.OUT_LEFT) > 0)
        {
          y = y1 + (y2 - y1) * (xmin - x1)/(x2 - x1);
          x = xmin;
        }
        //Now we move outside point to intersection point to clip
        //and get ready for next pass.
        if (outcodeOut == outcode0)
        {
          x1 = x;
          y1 = y;
          outcode0 = r.outcode(x1,y1);
        }
        else
        {
          x2 = x;
          y2 = y;
          outcode1 = r.outcode(x2,y2);
        }
      }
      if (hhh++>=5) {
        done=true;
        accept=true;
      }
    } while (!done);

    return accept ? new Line2D.Double(x1,y1,x2,y2) : null;
  }
}
