package gui;

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Stroke;
import java.awt.BasicStroke;
import java.awt.geom.Line2D;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.event.MouseEvent;
import java.awt.RenderingHints;
import java.awt.Image;
import java.awt.image.*;
import java.awt.geom.*;
import java.util.*;

import data.*;

public class VectorCanvas extends Canvas {
  private static final long serialVersionUID = 6093896460810726899L;
  
  private boolean init = true;
  ZoomAndPanListener zoomAndPanListener;

  public VectorCanvas(int width, int height) {
    setBackground(Color.WHITE);
    setPreferredSize(new Dimension(width, height));
    zoomAndPanListener = new ZoomAndPanListener(new ZoomAndPanAction() {
      public void mousePressed(MouseEvent e){}
      public void repaint(){Holder.INSTANCE.window.repaint();}
      public void setZoom(double level){if (Holder.INSTANCE!=null && Holder.INSTANCE.window!=null) Holder.INSTANCE.window.repaint();}
    }, 1.12);
//    zoomAndPanListener.setZoomLevel(1);
    this.addMouseListener(zoomAndPanListener);
    this.addMouseMotionListener(zoomAndPanListener);
    this.addMouseWheelListener(zoomAndPanListener);
  }

  private Image offscreen;

  @Override
  public void paint(Graphics g2) {
    Graphics2D g = (Graphics2D) g2;

    if (offscreen==null || offscreen.getWidth(null)!=getWidth() || offscreen.getHeight(null)!=getHeight()) {
      System.out.println("Creating offscreen image "+getSize());
//      offscreen = createVolatileImage(getWidth(), getHeight());
      offscreen = createImage(getWidth(), getHeight());
    }
    Graphics2D oldg = g;
    g = (Graphics2D)offscreen.getGraphics();
    g.clearRect(0,0,getWidth(),getHeight());
    g.setClip(oldg.getClipBounds());
    g.setTransform(oldg.getTransform());
    g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

    if (init) {
      // Initialize the viewport by moving the origin to the center of the
      // window and invert the y-axis to point upwards.
      init = false;
      Dimension d = getSize();    
      g.translate(0, d.height);
      g.scale(1, -.5);
      double s=4;
      g.scale(s,s);
      // Save the viewport to be updated by the ZoomAndPanListener
      zoomAndPanListener.setCoordTransform(g.getTransform());
    } else {
      // Restore the viewport after it was updated by the ZoomAndPanListener
      g.setTransform(zoomAndPanListener.getCoordTransform());
    }

//    System.out.println("Clip bounds: "+g.getClipBounds());

    //    g.setStroke(new BasicStroke((float)g.getTransform().getScaleX()));

    AffineTransform at = zoomAndPanListener.getCoordTransform();
    g.setTransform(new AffineTransform());

    Stroke oldStroke = g.getStroke();
    g.setStroke(new BasicStroke(3.f));

    // Draw axes for some clarity
    //   g.setStroke(new BasicStroke(1.1f));
    g.setColor(Color.BLUE);
    Util.drawLine(g, -5000, 0, 5000, 0, at);
    Util.drawLine(g, 0, -5000, 0, 5000, at);

    // Draw zoom level
    g.setColor(Color.GREEN);
    int xAxis = (int) (Holder.INSTANCE.getZoomFactor());
    Util.drawLine(g, xAxis, -5000, xAxis, 5000, at);

    g.setStroke(oldStroke);


    drawChosenLine(g, at);

    drawMedianLine(g, at);

    g.setColor(Color.BLACK);
    int idx = 1;
    for(Label label: Holder.INSTANCE.labels) {
//      if(label.visible) {
      label.paintDualLine(g, idx, at);
   // g.drawString(label.text, xAxis, (int) label.y);
      idx++;
//      }
    }

    g.dispose();
    oldg.drawImage(offscreen, 0, 0, null);
  }
  public void update(Graphics g) {
    paint(g);
  }

  void drawChosenLine(Graphics2D g, AffineTransform at) {
    /*
    if (zoomAndPanListener.lastMousePoint==null) return;
    if (!Holder.INSTANCE.dynamicClustering) return;
    try {
      Point2D.Double p = new Point2D.Double();
      at.inverseTransform(zoomAndPanListener.lastMousePoint, p);
//      Point2D p = zoomAndPanListener.transformPoint(zoomAndPanListener.lastMousePoint);
      if (p.x<0) return;
      int idx=1;
      Label bl=null;
      int bi=-1;
      double bd=1e100;
      for(Label l : Holder.INSTANCE.labels) {
        double y = l.y - p.x*idx;
        double d = Math.abs(y-p.y) / Math.sqrt(1 + l.y*l.y);
        if (d<bd) {
          bd=d;
          bl=l;
          bi=idx-1;
        }
        ++idx;
      }
      bd *= at.getScaleX();
      if (bd>.5) return;
//      System.out.println(bd+" "+bl);

      Iterable<LabelPositioner.ClusterInfo> clusters = Holder.INSTANCE.positioner.clustering.floorEntry(p.x).getValue();
      LabelPositioner.ClusterInfo info=null;
      for(LabelPositioner.ClusterInfo i : clusters) {
        if (i.points.contains(bi)) {
//          System.out.println("Chosen: "+i+" "+bi);
          info = i;
          break;
        }
      }
      g.setColor(Color.yellow);
      Stroke oldStroke = g.getStroke();
      g.setStroke(new BasicStroke(7));
      for(int i : info.points) {
        Label l = Holder.INSTANCE.positioner.points.get(i);
        l.paintDualLine(g, i+1, at);
      }
      g.setStroke(oldStroke);
      g.setColor(Color.black);
    } catch(NoninvertibleTransformException e) {
      e.printStackTrace();
    }
    */
  }

  void drawMedianLine(Graphics2D g, AffineTransform at) {
    Stroke oldStroke = g.getStroke();
    g.setStroke(new BasicStroke(4));
    g.setColor(Color.RED);

    List<Line2D.Double> lst;
    if (Holder.INSTANCE.dynamicClustering)
      lst = Holder.INSTANCE.getClusterLines();
    else
      lst = genMedianLine();

    for(Line2D.Double l : lst) {
      Point2D s = at.transform(new Point2D.Double(l.x1, l.y1), null);
      Point2D e = at.transform(new Point2D.Double(l.x2, l.y2), null);
      Util.drawLine(g, s, e);
    }

    g.setStroke(oldStroke);
  }
  static List<Line2D.Double> genMedianLine() {
    List<Line2D.Double> res = new ArrayList<Line2D.Double>();
      Point2D prev = null;

    Label[] labels = Holder.INSTANCE.labels.toArray(new Label[0]);
    if (labels.length%2==0) {
      for(Map.Entry<Double,Double> e : Holder.INSTANCE.getHeights().entrySet()) {
        Point2D p = new Point2D.Double(e.getKey(), e.getValue());
        if (prev!=null) {
          res.add(new Line2D.Double(prev.getX(), prev.getY(), p.getX(), p.getY()));
        }
        prev = p;
      }
    } else {
      int pm=0;
      for(Map.Entry<Double,Integer> e : Holder.INSTANCE.getMedianLine().entrySet()) {
        double x = e.getKey();
        x = Math.max(x,0);
        int m = e.getValue();
        double y = labels[m].y - x*(m+1);
        Point2D p = new Point2D.Double(x,y);
        if (prev!=null) {
          res.add(new Line2D.Double(prev.getX(), prev.getY(), p.getX(), p.getY()));
        }
        pm=m;
        prev = p;
      }
      double x = 5000;
      Point2D p = new Point2D.Double(x, labels[pm].y-x*(pm+1));
      res.add(new Line2D.Double(prev.getX(), prev.getY(), p.getX(), p.getY()));
    }
    return res;
  }
}
