package gui;

import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;

import javax.swing.JFrame;
import javax.swing.JOptionPane;

import data.Holder;
import data.Label;

/*
 * Copy&Paste coding from http://forums.sun.com/thread.jspa?threadID=5397418
 */

public class ZoomAndPanListener implements MouseListener, MouseMotionListener,
    MouseWheelListener {
  private boolean allowPan = true;

  private ZoomAndPanAction action;
  
//  public static final int MIN_ZOOM_LEVEL = 0;
  public static final int MIN_ZOOM_LEVEL = -80;
  public static final int MAX_ZOOM_LEVEL = 80;

  private double zoomMultiplicationFactor;
  private double zoomLevel = 0;
  private double scale = 1.0;
  
  private Point dragStartScreen;
  private AffineTransform coordTransform = new AffineTransform();

  Point lastMousePoint = null;

  public ZoomAndPanListener(ZoomAndPanAction action, double factor) {
    this.action = action;
    this.zoomMultiplicationFactor = factor;
  }

  public void mouseClicked(MouseEvent e) {
  }

  // create new label on right mouse click, otherwise pan
  public void mousePressed(MouseEvent e) {
    if (e.getButton() == MouseEvent.BUTTON1) {
      dragStartScreen = e.getPoint();
    } else action.mousePressed(e);
  }

  public void mouseReleased(MouseEvent e) {
    // moveCamera(e);
  }

  public void mouseEntered(MouseEvent e) {
    lastMousePoint = e.getPoint();
    action.repaint();
  }

  public void mouseExited(MouseEvent e) {
    lastMousePoint = null;
    action.repaint();
  }

  public void mouseMoved(MouseEvent e) {
    lastMousePoint = e.getPoint();
    action.repaint();
  }

  public void mouseDragged(MouseEvent e) {
    if (dragStartScreen!=null)
      moveCamera(e);
    dragStartScreen = e.getPoint();
    lastMousePoint = e.getPoint();
    action.repaint();
  }

  public void mouseWheelMoved(MouseWheelEvent e) {
    // System.out.println("============= Zoom camera ============");
    zoomCamera(e);
  }

  private void moveCamera(MouseEvent e) {
    // System.out.println("============= Move camera ============");
    if(!allowPan) {
      return;
    }
    
    try {
      Point dragEndScreen = e.getPoint();
      Point2D.Float dragStart = transformPoint(dragStartScreen);
      Point2D.Float dragEnd = transformPoint(dragEndScreen);
      double dx = dragEnd.getX() - dragStart.getX();
      double dy = dragEnd.getY() - dragStart.getY();
      coordTransform.translate(dx, dy);
      action.repaint();
    } catch (NoninvertibleTransformException ex) {
      ex.printStackTrace();
    }
  }

  private void zoomCamera(MouseWheelEvent e) {
    int wheelRotation = e.getWheelRotation();
    Point p = e.getPoint();
    if(wheelRotation > 0) {
      scale *= (1 / zoomMultiplicationFactor);
    } else {
      scale *= (zoomMultiplicationFactor);
    }
    
//    System.out.println(scale);
    
    try {
      if (wheelRotation > 0) {
        if (zoomLevel < MAX_ZOOM_LEVEL) {
          zoomLevel += 1;
          Point2D p1 = transformPoint(p);
          coordTransform.scale(1 / zoomMultiplicationFactor,
              1 / zoomMultiplicationFactor);
          Point2D p2 = transformPoint(p);
          coordTransform
              .translate(p2.getX() - p1.getX(), p2.getY() - p1.getY());
        }
      } else {
        if (zoomLevel > MIN_ZOOM_LEVEL) {
          zoomLevel -= 1;
          Point2D p1 = transformPoint(p);
          coordTransform.scale(zoomMultiplicationFactor,
              zoomMultiplicationFactor);
          Point2D p2 = transformPoint(p);
          coordTransform
              .translate(p2.getX() - p1.getX(), p2.getY() - p1.getY());
        }
      }
      
      action.setZoom(zoomLevel);
    } catch (NoninvertibleTransformException ex) {
      ex.printStackTrace();
    }
    
//    System.out.println("Current zoom level: " + zoomLevel);
  }
  
  public double getCurrentZoomMultiplicationFactor() {
//    System.out.println("lol "+scale+" "+Math.pow(zoomMultiplicationFactor, zoomLevel)+" "+zoomLevel);
    return Math.pow(zoomMultiplicationFactor, zoomLevel);
  }

  public Point2D.Float transformPoint(Point p1)
      throws NoninvertibleTransformException {
    Point2D.Float r = new Point2D.Float();
    coordTransform.inverseTransform(p1, r);
    return r;
  }

  public double getZoomLevel() {
    return zoomLevel;
  }

  public void setZoomLevel(double zoomLevel) {
    this.zoomLevel = zoomLevel;
    scale = getCurrentZoomMultiplicationFactor();
    action.setZoom(zoomLevel);
    coordTransform = AffineTransform.getTranslateInstance(coordTransform.getTranslateX(), coordTransform.getTranslateY());
    coordTransform.scale(scale, -scale);
  }

  public AffineTransform getCoordTransform() {
    return coordTransform;
  }

  public void setCoordTransform(AffineTransform coordTransform) {
    this.coordTransform = coordTransform;
//    dragStartScreen = lastMousePoint;
//    dragEndScreen = lastMousePoint;
    dragStartScreen = null;
  }

  public void resetView(double x0, double y0, double s, double h) {
    System.out.println("setting view to "+x0+" "+y0+" "+s);
    scale = s;
    double oldZ = zoomLevel;
    zoomLevel = -Math.log(scale) / Math.log(zoomMultiplicationFactor);
//    zoomLevel /= 15;
    System.out.println("Zoom level "+oldZ+" -> "+zoomLevel);
//    coordTransform = AffineTransform.getTranslateInstance(x0, y0);
    coordTransform = new AffineTransform();
    coordTransform.translate(5, h-5);
    coordTransform.scale(scale, -scale);
    action.setZoom(zoomLevel);
//    coordTransform.translate(x0, y0);
//    setZoomLevel((int)(Math.log(s)/Math.log(zoomMultiplicationFactor)) - 1);
  }
}
